<?php

namespace UnicaenDbImport\Controller;

use UnicaenDbImport\Domain\ImportResult;
use UnicaenDbImport\Domain\Synchro;
use UnicaenDbImport\Domain\SynchroResult;
use UnicaenDbImport\Entity\Db\Service\ImportLog\ImportLogServiceAwareTrait;
use UnicaenDbImport\Service\Traits\SynchroServiceAwareTrait;
use Laminas\Http\Request;
use Laminas\Http\Response;
use Laminas\Mvc\Controller\AbstractActionController;
use Laminas\View\Model\JsonModel;

/**
 * @method \Laminas\Mvc\Plugin\FlashMessenger\FlashMessenger flashMessenger()
 */
class SynchroController extends AbstractActionController
{
    use SynchroServiceAwareTrait;
    use ImportLogServiceAwareTrait;

    /**
     * Listage des synchros.
     *
     * @return array
     */
    public function indexAction(): array
    {
        $synchros = $this->synchroService->getSynchros();

        $importLogs = $diffCounts = [];
        foreach ($synchros as $synchro) {
            $importLog = $this->importLogService->findLastLogForSynchro($synchro);
            $importLogs[$synchro->getName()] = $importLog;
            $diffCounts[$synchro->getName()] = $this->synchroService->fetchSynchroDiffCount($synchro);
        }

        return [
            'synchros' => $synchros,
            'importLogs' => $importLogs,
            'diffCounts' => $diffCounts,
        ];
    }

    /**
     * Détails d'une synchro.
     *
     * @return array
     */
    public function voirAction(): array
    {
        $name = $this->params('name');
        $synchro = $this->synchroService->getSynchroByName($name);
        $diffCount = $this->synchroService->fetchSynchroDiffCount($synchro);
        $importLog = $this->importLogService->findLastLogForSynchro($synchro);

        return [
            'synchro' => $synchro,
            'diffCount' => $diffCount,
            'limit' => 30,
            'importLog' => $importLog,
        ];
    }

    /**
     * Différentiel d'une synchro.
     *
     * @return array
     */
    public function diffAction(): array
    {
        $name = $this->params('name');
        $synchro = $this->synchroService->getSynchroByName($name);

        $diff = $this->synchroService->fetchSynchroDiff($synchro);
        $count = $this->synchroService->fetchSynchroDiffCount($synchro);

        return [
            'synchro' => $synchro,
            'diff' => $diff,
            'limit' => 30,
            'count' => $count,
        ];
    }

    /**
     * @return \Laminas\View\Model\JsonModel
     */
    public function diffCountAction(): JsonModel
    {
        $name = $this->params('name');
        $synchro = $this->synchroService->getSynchroByName($name);

        return new JsonModel([
            'count' => $this->synchroService->fetchSynchroDiffCount($synchro),
        ]);
    }

    /**
     * @return \Laminas\Http\Response
     * @throws \Exception
     */
    public function lancerAction(): Response
    {
        $name = $this->params('name');
        $synchro = $this->synchroService->getSynchroByName($name);

        $result = $this->synchroService->runSynchro($synchro);

        $startDate = $result->getStartDate();
        $endDate = $result->getEndDate();
        $diffDate = date_diff($startDate, $endDate);
        $duration = $diffDate->format('%i min %s sec.');

        if ($exception = $result->getFailure()) {
            throw $exception;
        }

        $error = "La synchro a rencontré un problème ($duration).";
        $success = "Synchro réalisée avec succès ($duration).";
        $info = str_replace(PHP_EOL, '<br>', $result->toString(true));

        if ($result->hasExceptionInResults()) {
            $this->flashMessenger()->addErrorMessage($error);
        } else {
            $this->flashMessenger()->addSuccessMessage($success);
        }
        $this->flashMessenger()->addInfoMessage($info);

        $redirect = $this->params()->fromQuery('redirect') ?:
            $this->url()->fromRoute('unicaen-db-import/synchro/diff', ['name' => $synchro->getName()]);

        return $this->redirect()->toUrl($redirect);
    }

    /**
     * @return \Laminas\View\Model\JsonModel
     */
    public function lastLogAction(): JsonModel
    {
        $name = $this->params('name');
        $synchro = $this->synchroService->getSynchroByName($name);
        $importLog = $this->importLogService->findLastLogForSynchro($synchro);

        if ($importLog === null) {
            return new JsonModel(null);
        }

        return new JsonModel([
            'endedOn' => $importLog->getEndedOnToString(),
            'isSuccess' => $importLog->isSuccess(),
        ]);
    }
}