<?php

namespace UnicaenDbImport\Domain;

use Doctrine\DBAL\Connection;
use UnicaenDbImport\Filter\ColumnName\ColumnNameFilterInterface;
use UnicaenDbImport\Filter\ColumnValue\ColumnValueFilterInterface;

interface DestinationInterface
{
    const LOG_TABLE_DEFAULT_NAME = 'import_log';

    const HISTO_COLUMNS = ['CREATED_ON', 'UPDATED_ON', 'DELETED_ON'];

    /**
     * Version textuelle de cette destination.
     *
     * @return string
     */
    public function __toString(): string;

    /**
     * Retourne le petit nom de cette destination.
     *
     * @return string
     */
    public function getName(): string;

    /**
     * Retourne le nom de la table de destination.
     *
     * @return string
     */
    public function getTable(): string;

    /**
     * Retourne le filtre SQL éventuel permettant de restreindre les données destinations concernées.
     *
     * @return string|null
     */
    public function getWhere(): ?string;

    /**
     * Retourne l'instance de la connexion Doctrine vers cette destination.
     *
     * @return Connection
     */
    public function getConnection(): Connection;

    /**
     * Retourne le nom de la colonne commune existant obligatoirement dans la table/vue source
     * ET la table destination.
     *
     * Dans les tables source et destination, cette colonne est celle d'un identifiant unique permettant
     * de rapprocher les enregistrements source et destination.
     *
     * @return string Exemple : 'code'
     */
    public function getSourceCodeColumn(): string;

    /**
     * Retourne les noms des colonnes de la table destination concernées par l'import.
     *
     * Ne doit pas inclure la colonne commune retournée par {@link getSourceCodeColumn}.
     *
     * @return string[] Exemple : ['libelle', 'debut_validite', 'fin_validite']
     */
    public function getColumns(): array;

    /**
     * Retourne l'instance de filtre permettant de transformer une veleur de colonne/attribut source
     * en une valeur de colonne destination.
     *
     * @return \UnicaenDbImport\Filter\ColumnValue\ColumnValueFilterInterface|null
     */
    public function getColumnValueFilter(): ?ColumnValueFilterInterface;


    /**
     * Retourne le nom éventuel de la séquence permettant de générer les valeurs dans la colonne ID (clé primaire).
     * Si la valeur retournée est `null`, la séquence par défaut '{table}_ID_SEQ' sera utilisée.
     * Si la valeur retournée est `false`, la colonne devra avoir un type auto-incrémenté
     * pour que l'insertion de données fonctionne.
     *
     * @return string|null|false
     */
    public function getIdColumnSequence();

    /**
     * Retourne le nom de la table intermédiaire à créer dans la base de données de destination
     * lorsque l'utilisation d'une table intermédiaire est nécessaire (cf {@link requiresIntermediateTable}).
     *
     * @return string
     */
    public function getIntermediateTable(): string;

    /**
     * Indique si la suppression automatique de la table intermédiaire est activée.
     *
     * @return bool
     */
    public function getIntermediateTableAutoDrop(): bool;

    /**
     * Retourne le nom de la table des logs.
     *
     * @return string
     */
    public function getLogTable(): string;

}