<?php

namespace UnicaenDbImport\Domain;

use DateTime;
use Exception;
use Ramsey\Uuid\Uuid;
use UnicaenApp\Exception\RuntimeException;

abstract class Result implements ResultInterface
{
    /**
     * @var Exception
     */
    private $failure;

    /**
     * @var array [operation => int|Exception]
     */
    private $resultsByOperation = [];

    /**
     * @var bool
     */
    private $hasExceptionInResults = false;

    /**
     * @var DateTime
     */
    private $startDate;

    /**
     * @var DateTime
     */
    private $endDate;

    /**
     * @var string
     */
    private $hash;

    /**
     * @param array $results [string => int|Exception]
     */
    public function setResultsByOperation(array $results)
    {
        $this->hasExceptionInResults = false;

        foreach ($results as $operation => $result) {
            /** @var int|Exception $result */
            $this->setResultForOperation($result, $operation);
        }
    }

    /**
     * @param int|Exception $result
     * @param string $operation Ex: {@see Operation::OPERATION_INSERT}
     */
    public function setResultForOperation($result, $operation)
    {
        $this->resultsByOperation[$operation] = $result;

        if ($result instanceof Exception) {
            $this->hasExceptionInResults = true;
        }
    }

    /**
     * @return Exception
     */
    public function getFailure()
    {
        return $this->failure;
    }

    /**
     * @param Exception $failure
     */
    public function setFailure(Exception $failure = null)
    {
        $this->failure = $failure;
    }

    /**
     * @return bool
     */
    public function hasExceptionInResults(): bool
    {
        return $this->hasExceptionInResults;
    }

    /**
     * @return DateTime
     */
    public function getStartDate(): DateTime
    {
        return $this->startDate;
    }

    /**
     * @param DateTime $startDate
     */
    public function setStartDate(DateTime $startDate = null)
    {
        $this->startDate = $startDate ?: date_create();
    }

    /**
     * @return DateTime
     */
    public function getEndDate(): DateTime
    {
        return $this->endDate;
    }

    /**
     * @param DateTime $endDate
     */
    public function setEndDate(DateTime $endDate = null)
    {
        $this->endDate = $endDate ?: date_create();
    }

    /**
     * @param bool $includeDetails
     * @return string
     */
    public function toString($includeDetails = false)
    {
        if (count($this->resultsByOperation) === 0) {
            return "Aucune instruction exécutée.";
        }

        $str = '';
        foreach ($this->resultsByOperation as $operation => $result) {
            $str .= "# " . sprintf('%-12s', $operation . " : ");
            if (is_array($result)) {
                $str .= $result['count'] . " instructions exécutées.";
            } elseif ($result instanceof Exception){
                $str .= $result->getMessage();
                if ($includeDetails) {
                    $str .= PHP_EOL . $result->getTraceAsString();
                }
            } elseif (is_int($result)) {
                $str .= $result . " enregistrement(s).";
            } else {
                throw new RuntimeException("Type de résultat spécifié inattendu : " . gettype($result));
            }
            $str .= PHP_EOL;
        }

        return $str;
    }

    /**
     * @return string
     * @throws \Exception
     */
    public function getHash(): string
    {
        if ($this->hash === null) {
            $this->hash = Uuid::uuid1();
        }

        return $this->hash;
    }

    /**
     * @return string
     */
    public function __toString()
    {
        return $this->toString();
    }
}