<?php

namespace UnicaenDbImport\Service;

use Doctrine\DBAL\Connection as DbConnection;
use stdClass;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\Connection\ApiConnection;
use UnicaenDbImport\Domain\DestinationInterface;
use UnicaenDbImport\Domain\Exception\ConnectionException;
use UnicaenDbImport\Domain\ImportResult;
use UnicaenDbImport\Domain\ResultInterface;
use UnicaenDbImport\Domain\SourceInterface;
use UnicaenDbImport\Domain\SynchroResult;
use Laminas\Log\Logger;
use Laminas\Log\LoggerInterface;
use Laminas\Log\Writer\Stream;

class FacadeService
{
    /**
     * @var DatabaseService
     */
    private $databaseService;

    /**
     * @var ApiService
     */
    private $apiService;

    /**
     * @var SourceInterface $source
     */
    protected $source;

    /**
     * @var DestinationInterface $destination
     */
    protected $destination;

    /**
     * @var bool
     */
    protected $useImportObserv = false;

    /**
     * @var bool
     */
    protected $importLogTableCreated = false;

    /**
     * FacadeService constructor.
     *
     * @param DatabaseService $databaseService
     * @param ApiService      $apiService
     */
    public function __construct(DatabaseService $databaseService, ApiService $apiService)
    {
        $this->databaseService = $databaseService;
        $this->apiService = $apiService;
    }

    /**
     * @param SourceInterface $source
     */
    public function setSource(SourceInterface $source): void
    {
        $this->source = $source;
        $this->databaseService->setSource($this->source);
        $this->apiService->setSource($this->source);
    }

    /**
     * @param DestinationInterface $destination
     */
    public function setDestination(DestinationInterface $destination): void
    {
        $this->destination = $destination;
        $this->databaseService->setDestination($this->destination);
        $this->apiService->setDestination($this->destination);
    }

    /**
     * @param bool $useImportObserv
     */
    public function setUseImportObserv(bool $useImportObserv)
    {
        $this->useImportObserv = $useImportObserv;
    }

    public function validateSource()
    {
        $connection = $this->source->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                if (! $this->requiresIntermediateTable()) {
                    $this->databaseService->validateSource();
                }
                break;
            case $connection instanceof ApiConnection:
                // Ex:
                // $this->apiService->validateSourceApi();
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    public function validateDestination()
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $this->databaseService->validateDestination();
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    public function createImportLog()
    {
        if ($this->importLogTableCreated) {
            return;
        }

        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $this->databaseService->createImportLogTableIfNotExists();
                $this->importLogTableCreated = true;
                break;
            default:
                throw ConnectionException::unexpected($connection);
        }
    }

    /**
     *
     */
    public function clearDestination()
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $this->databaseService->truncateDestinationTable();
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    /**
     * @return ImportResult
     */
    public function importToDestination(): ImportResult
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $count = $this->databaseService->populateDestinationTableFromSource();

                $result = new ImportResult();
                $result->setDestinationTablePopulateResult($count);

                return $result;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    /**
     * @return array
     */
    public function fetchSynchroDiff(): array
    {
        $this->recreateDiffView();

        return $this->databaseService->fetchDiffView();
    }

    /**
     * @return int
     */
    public function fetchSynchroDiffCount(): int
    {
        $this->recreateDiffView();

        return $this->databaseService->fetchDiffViewCount();
    }

    /**
     * @return SynchroResult
     */
    public function synchronizeDestination(): SynchroResult
    {
        $intermediateTable = null;

        if ($this->requiresIntermediateTable()) {
            $this->fetchSource();

            $intermediateTable = $this->destination->getIntermediateTable();
            if ($this->destination->getIntermediateTableAutoDrop()) {
                $this->databaseService->dropIntermediateTable($intermediateTable);
            } else {
                $this->databaseService->checkIntermediateTableNotExists($intermediateTable);
            }
            $this->databaseService->createIntermediateTable($intermediateTable);
            $this->databaseService->populateTableFromSource($intermediateTable, false);
        }

        $this->recreateDiffView();
        if ($this->useImportObserv) {
            $this->databaseService->populateImportObservResultTable();
        }
        
        $resultsByOperation = $this->databaseService->updateDestination();
        
        $result = new SynchroResult();
        $result->setResultsByOperation($resultsByOperation);
        if ($result->hasExceptionInResults()) {
            return $result;
        }

        if ($this->requiresIntermediateTable()) {
            $this->databaseService->dropIntermediateTable($intermediateTable);
        }

        return $result;
    }

    protected function recreateDiffView()
    {
        $intermediateTable = null;
        if ($this->requiresIntermediateTable()) {
            $intermediateTable = $this->destination->getIntermediateTable();
        }

        $this->databaseService->recreateDiffView($intermediateTable);
    }

    /**
     * @return bool
     */
    private function requiresIntermediateTable(): bool
    {
        $isSameConnectionForSourceAndDestination = $this->source->getConnection() === $this->destination->getConnection();

        return ! $isSameConnectionForSourceAndDestination;
    }

    /**
     * Requête la Source pour obtenir les données sources.
     * Ces données sont ensuite disponibles dans la Source en question.
     *
     * @see SourceInterface::getData()
     */
    public function fetchSource()
    {
        if (count($this->source->getData()) === 0) {

            /** @var ApiConnection|DbConnection $connection */
            $connection = $this->source->getConnection();

            switch (true) {
                case $connection instanceof ApiConnection:
                    $this->apiService->setLogger($this->createLogger());
                    $this->apiService->setConfig($connection->toArray());
                    $data = $this->apiService->fetchAll($connection, $this->source);
                    break;

                case $connection instanceof DbConnection:
                    $data = $this->databaseService->fetchSource();
                    break;

                default:
                    throw new RuntimeException("Type de connexion source inattendue");
            }

            $this->source->setData($data);
        }
    }

    /**
     * Requête la Source pour déterminer la liste des colonnes.
     *
     * @see SourceInterface::getColumns()
     * @see SourceInterface::getData()
     */
    public function discoverSourceColumns()
    {
        $data = $this->source->getData();

        if (count($data) === 0) {

            /** @var ApiConnection|DbConnection $connection */
            $connection = $this->source->getConnection();

            switch (true) {
                case $connection instanceof ApiConnection:
                    $this->apiService->setLogger($this->createLogger());
                    $this->apiService->setConfig($connection->toArray());
                    $data = $this->apiService->fetchFirstRow($connection, $this->source);
                    break;

                case $connection instanceof DbConnection:
                    $firstRow = $this->databaseService->fetchSourceFirstRow();
                    if ($firstRow === null) {
                        throw new RuntimeException(
                            "Impossible de découvrir la liste des colonnes à partir des données sources car celles-ci sont vides");
                    }
                    $data = [ $firstRow ];
                    break;

                default:
                    throw new RuntimeException("Type de connexion source inattendue");
            }
        }

        $columns = $this->extractColumnsFromSourceFetchResults($data);

        $this->source->setColumns($columns);
    }

    /**
     * @param array $rows
     * @return string[]
     */
    private function extractColumnsFromSourceFetchResults(array $rows): array
    {
        $first = current($rows);

        if ($first instanceof stdClass) {
            $columns = array_keys(get_object_vars($first));
        } elseif (is_array($first)) {
            $columns = array_keys($first);
        } else {
            throw new RuntimeException(
                "Impossible de déterminer la liste des colonnes à partir des données sources car leur type est inconnu (ni array ni stdClass)");
        }

        return array_diff($columns, ['source_id', 'SOURCE_ID']);
    }

    /**
     * @return LoggerInterface
     */
    private function createLogger()
    {
        $logger = new Logger();
        $logger->addWriter(new Stream('php://output'));

        return $logger;
    }

    /**
     * @param ResultInterface $result
     */
    public function saveResultToImportLog(ResultInterface $result)
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $this->databaseService->saveResultToLogTable($result);
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }
}