<?php

namespace UnicaenDbImport\Service;

use Exception;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\Config\Config;
use UnicaenDbImport\Domain\Exception\NotFoundException;
use UnicaenDbImport\Domain\ImportInterface;
use UnicaenDbImport\Domain\ImportResult;

class ImportService
{
    /**
     * @var Config
     */
    private $config;

    /**
     * @var FacadeService
     */
    private $facadeService;

    /**
     * ImportService constructor.
     *
     * @param FacadeService $facadeService
     * @param Config        $config
     */
    public function __construct(
        FacadeService $facadeService,
        Config $config)
    {
        $this->facadeService = $facadeService;
        $this->config = $config;
    }

    /**
     * Retourne tous les imports (bruts) issus de la config.
     *
     * @return ImportInterface[]
     */
    public function getImports(): array
    {
        return $this->config->getImports();
    }

    /**
     * Lance tous les imports.
     *
     * @return ImportResult[] Agrégat des résultats des l'exécutions des imports
     */
    public function runAllImports(): array
    {
        $imports = $this->getImports();
        $results = [];

        foreach ($imports as $import) {
            $results[$import->getName()] = $this->runImport($import);
        }

        return $results;
    }

    /**
     * Retourne l'import dont le nom est spécifié.
     *
     * @param string $name Nom de l'import (name)
     * @return ImportInterface Import trouvé
     *
     * @throws NotFoundException Import introuvable
     */
    public function getImportByName(string $name): ImportInterface
    {
        $import = $this->config->getImport($name);
        if ($import === null) {
            throw NotFoundException::importByName($name);
        }

        return $import;
    }

    /**
     * Lance l'import spécifié (import brut).
     *
     * @param ImportInterface $import
     * @return ImportResult Résultat de l'exécution de l'import
     */
    public function runImport(ImportInterface $import): ImportResult
    {
        $source = $import->getSource();
        $destination = $import->getDestination();
        $connection = $destination->getConnection();

        $startDate = date_create('now');

        $this->facadeService->setSource($source);
        $this->facadeService->setDestination($destination);

        try {
            $this->facadeService->createImportLog();
            $this->facadeService->validateDestination();
            $this->facadeService->fetchSource();
            $this->facadeService->discoverSourceColumns();

            $connection->beginTransaction();
            try {
                $this->facadeService->clearDestination();
                $result = $this->facadeService->importToDestination();
                $connection->commit();
            } catch (Exception $e) {
                $connection->rollBack();
                throw new RuntimeException("Erreur rencontrée. Rollback effectué.", null, $e);
            }
        } catch (Exception $e) {
            $result = new ImportResult();
            $result->setFailure($e);
        }

        $result->setImport($import);
        $result->setStartDate($startDate);
        $result->setEndDate();

        $this->facadeService->saveResultToImportLog($result);

        return $result;
    }
}