<?php

namespace UnicaenDbImport\Service;

use Exception;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\Config\Config;
use UnicaenDbImport\Domain\Exception\NotFoundException;
use UnicaenDbImport\Domain\ImportInterface;
use UnicaenDbImport\Domain\SynchroInterface;
use UnicaenDbImport\Domain\SynchroResult;

class SynchroService
{
    /**
     * @var Config
     */
    private $config;

    /**
     * @var FacadeService
     */
    private $facadeService;

    /**
     * SynchroService constructor.
     *
     * @param FacadeService $facadeService
     * @param Config        $config
     */
    public function __construct(FacadeService $facadeService, Config $config)
    {
        $this->facadeService = $facadeService;
        $this->config = $config;
    }


    /**
     * Retourne toutes les synchros issues de la config.
     *
     * @return SynchroInterface[]
     */
    public function getSynchros()
    {
        return $this->config->getSynchros();
    }

    /**
     * Lance tous les synchros.
     *
     * @return SynchroResult[] Agrégat des résultats des l'exécutions des synchros
     */
    public function runAllSynchros()
    {
        $synchros = $this->getSynchros();
        $results = [];

        foreach ($synchros as $synchro) {
            $results[$synchro->getName()] = $this->runSynchro($synchro);
        }

        return $results;
    }

    /**
     * Retourne la synchro dont le nom est spécifié.
     *
     * @param string $name Nom de la synchro (name)
     * @return SynchroInterface Synchro trouvée
     *
     * @throws NotFoundException Synchro introuvable
     */
    public function getSynchroByName($name)
    {
        $synchro = $this->config->getSynchro($name);
        if ($synchro === null) {
            throw NotFoundException::synchroByName($name);
        }

        return $synchro;
    }

    /**
     * Lance la synchro spécifiée.
     *
     * @param SynchroInterface $synchro
     * @return SynchroResult Résultat de l'exécution de la synchro
     */
    public function runSynchro(SynchroInterface $synchro)
    {
        $source = $synchro->getSource();
        $destination = $synchro->getDestination();
        $connection = $destination->getConnection();

        $this->facadeService->setSource($source);
        $this->facadeService->setDestination($destination);

        $intermediateTable = null;

        $startDate = date_create('now');

        $exception = null;
        try {
            $this->facadeService->createImportLog();
            $this->facadeService->validateDestination();
            $this->facadeService->validateSource();
            $this->facadeService->discoverSourceColumns();

            $connection->beginTransaction();
            try {
                $result = $this->facadeService->synchronizeDestination();
                $connection->commit();
            } catch (Exception $e) {
                $connection->rollBack();
                throw new RuntimeException("Erreur rencontrée. Rollback effectué.", null, $e);
            }
        } catch (Exception $e) {
            $result = new SynchroResult();
            $result->setFailure($e);
        }

        $result->setSynchro($synchro);
        $result->setStartDate($startDate);
        $result->setEndDate();

        $this->facadeService->saveResultToImportLog($result);

        return $result;
    }

    /**
     * @param \UnicaenDbImport\Domain\SynchroInterface $synchro
     * @return array
     */
    public function fetchSynchroDiff(SynchroInterface $synchro): array
    {
        $source = $synchro->getSource();
        $destination = $synchro->getDestination();

        $this->facadeService->setSource($source);
        $this->facadeService->setDestination($destination);

        try {
            $this->facadeService->validateDestination();
            $this->facadeService->validateSource();
            $this->facadeService->discoverSourceColumns();

            return $this->facadeService->fetchSynchroDiff();

        } catch (Exception $e) {
            throw new RuntimeException("Erreur rencontrée.", null, $e);
        }
    }

    /**
     * @param \UnicaenDbImport\Domain\SynchroInterface $synchro
     * @return int
     */
    public function fetchSynchroDiffCount(SynchroInterface $synchro): int
    {
        $source = $synchro->getSource();
        $destination = $synchro->getDestination();

        $this->facadeService->setSource($source);
        $this->facadeService->setDestination($destination);

        try {
            $this->facadeService->validateDestination();
            $this->facadeService->validateSource();
            $this->facadeService->discoverSourceColumns();

            return $this->facadeService->fetchSynchroDiffCount();

        } catch (Exception $e) {
            throw new RuntimeException("Erreur rencontrée.", null, $e);
        }
    }
}