<?php

namespace UnicaenDbImport\CodeGenerator\Helper;

use Doctrine\DBAL\DBALException;
use Doctrine\DBAL\Exception;
use InvalidArgumentException;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\CodeGenerator\Helper;

/**
 * Génération de code SQL pour tester l'existence, la conformité de table.
 *
 * Version commune à toutes les plateformes de bases de données.
 *
 * @author Unicaen
 */
abstract class TableValidationHelper extends Helper
{
    /**
     * @param string $tableName
     * @return string
     */
    abstract public function generateSQLForTableExistenceCheck($tableName);

    /**
     * @param string $sequenceName
     * @return string
     */
    abstract public function generateSQLForSequenceExistenceCheck($sequenceName);

    /**
     * @param array $result
     * @return bool
     * @throws \UnicaenDbImport\CodeGenerator\Helper\HelperException
     */
    public function convertTableExistenceCheckResultToBoolean(array $result)
    {
        if (! isset($result[0])) {
            throw new HelperException("Données fournies non exploitables");
        }

        $value = array_pop($result[0]); // NB: on ignore le nom de la colonne car la casse est incertaine

        return intval($value) === 1;
    }

    /**
     * @param string $sequenceName
     * @param array $result
     * @return null|RuntimeException
     */
    public function convertSequenceExistenceCheckResultToException(string $sequenceName, array $result)
    {
        $value = array_pop($result[0]); // NB: on ignore le nom de la colonne car la casse est incertaine

        if (intval($value) === 1) {
            return null;
        }

        $message = "Assurez-vous que la sequence '$sequenceName' existe : " . PHP_EOL;
        $message .= <<<EOT
CREATE SEQUENCE $sequenceName ;
EOT;

        return new RuntimeException($message);
    }

    /**
     * @param string $tableName
     * @param array  $columnsAndTypes
     * @return string
     */
    abstract public function generateSQLForColumnsValidation($tableName, array $columnsAndTypes);

    /**
     * @param string $tableName
     * @param array  $result
     * @param array  $expectedColumns
     * @return null|RuntimeException
     */
    public function convertColumnsValidationResultToException($tableName, array $result, array $expectedColumns)
    {
        if (count($result) === count($expectedColumns)) {
            return null;
        }

        $message = "Assurez-vous que les colonnes suivantes existent bien dans la table '$tableName' : ";
        $message .= implode(', ', $expectedColumns);

        return new RuntimeException($message);
    }

    /**
     * @param string $tableName
     * @param int $columnsCount
     * @return string
     */
    abstract public function generateSQLForIdColumnValidation(string $tableName, int &$columnsCount);

    /**
     * @param string $tableName
     * @param int $columnsCount
     * @return string
     */
    abstract public function generateSQLForHistoColumnsValidation(string $tableName, int &$columnsCount);

    /**
     * @param string $tableName
     * @param string $sourceCodeColumn
     * @param int $columCount
     * @return string
     */
    abstract public function generateSQLForSourceColumnsValidation(string $tableName, string $sourceCodeColumn, int &$columCount);

    /**
     * @param string $tableName
     * @param array $result
     * @param int $columnCount
     * @return null|RuntimeException
     */
    public function convertIdColumnValidationBadResultToException($tableName, array $result, int $columnCount)
    {
        if (count($result) === $columnCount) {
            return null;
        }

        $integer = $this->platform->getBigIntTypeDeclarationSQL([]);

        $message = "Assurez-vous que la colonne id existe dans la table '$tableName' : " . PHP_EOL;
        $message .= <<<EOT
ALTER TABLE $tableName ADD id $integer NOT NULL ;
EOT;

        return new RuntimeException($message);
    }

    /**
     * @param string $tableName
     * @param array $result
     * @param int $columnCount
     * @return null|RuntimeException
     * @throws \UnicaenDbImport\CodeGenerator\Helper\HelperException
     */
    public function convertHistoColumnsValidationBadResultToException($tableName, array $result, int $columnCount)
    {
        if (count($result) === $columnCount) {
            return null;
        }

        try {
            $timestamp = $this->platform->getDateTimeTypeDeclarationSQL([]);
        } catch (Exception $e) {
            throw new HelperException("Type non supporté par la plateforme de base de données !", null, $e);
        }
        $integer = $this->platform->getBigIntTypeDeclarationSQL([]);
        $now = $this->platform->getNowExpression();

        $createdOnCol = $this->config->getHistoColumnAliasForCreatedOn();
        $updatedOnCol = $this->config->getHistoColumnAliasForUpdatedOn();
        $deletedOnCol = $this->config->getHistoColumnAliasForDeletedOn();
        $createdByCol = $this->config->getHistoColumnAliasForCreatedBy();
        $updatedByCol = $this->config->getHistoColumnAliasForUpdatedBy();
        $deletedByCol = $this->config->getHistoColumnAliasForDeletedBy();

        $message = "Assurez-vous que les colonnes d'historique suivantes sont présentes dans la table '$tableName' : " . PHP_EOL;
        $message .= <<<EOT
ALTER TABLE $tableName ADD $createdOnCol $timestamp DEFAULT $now NOT NULL ;
ALTER TABLE $tableName ADD $updatedOnCol $timestamp ;
ALTER TABLE $tableName ADD $deletedOnCol $timestamp ;
ALTER TABLE $tableName ADD $createdByCol $integer NOT NULL ;
ALTER TABLE $tableName ADD $updatedByCol $integer ;
ALTER TABLE $tableName ADD $deletedByCol $integer ;
EOT;

        return new RuntimeException($message);
    }

    /**
     * @param string $tableName
     * @param string $sourceCodeColumn
     * @param array $result
     * @param int $columCount
     * @return null|RuntimeException
     */
    public function convertSourceColumnsValidationBadResultToException(string $tableName, string $sourceCodeColumn, array $result, int $columCount)
    {
        if (count($result) === $columCount) {
            return null;
        }

        $integer = $this->platform->getBigIntTypeDeclarationSQL([]);
        $varchar = $this->platform->getVarcharTypeDeclarationSQL(['length' => 100]);

        $message = "Assurez-vous que les colonnes suivantes sont présentes dans la table '$tableName' : " . PHP_EOL;
        $message .= <<<EOT
ALTER TABLE $tableName ADD source_id $integer NOT NULL ;
ALTER TABLE $tableName ADD $sourceCodeColumn $varchar ;
EOT;

        return new RuntimeException($message);
    }
}